import os
import argparse
import warnings
import random
import anndata as ad
import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import Counter
from tqdm import tqdm
import math
import scanpy as sc
from sklearn.metrics import accuracy_score
from sklearn.metrics.cluster import normalized_mutual_info_score

import numpy as np
import torch
import torch.cuda as cuda
from torch.utils.data import Dataset, DataLoader
from torch_geometric.nn import GCNConv, GATConv
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.inits import glorot, uniform
from torch_geometric.utils import softmax as Softmax
from torchmetrics.functional import pairwise_cosine_similarity
import anndata as ad
import h5py
import scipy
import scipy.io
from scipy import io
from scipy.io import mmwrite,mmread
from scipy.sparse import csc_matrix, csr_matrix

from bioservices import KEGG
from collections import Counter
from itertools import chain
import multiprocessing as mp
import matplotlib.pyplot as plt
from pandas.plotting import table
import seaborn as sns


# Ignore warnings
def ignore_warnings():
    warnings.filterwarnings("ignore")

    
# Set random seed
def set_random_seed(seed=0):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

    
# Argument parser
def parse_arguments(labsm=0.3, wd=0.1, lr=0.0005,
                   n_hid=104, nheads=8, nlayers=3, 
                   cell_size=30, neighbor=20, egrn=True, 
                   output_file='/users/PCON0022/duanmaoteng/Metastasis/GSE197177'):
    parser = argparse.ArgumentParser(description='Training GNN on gene cell graph')
    parser.add_argument('--labsm', type=float, default=labsm)  # The rate of Label Smoothing
    parser.add_argument('--wd', type=float, default=wd)  # The weight decay
    parser.add_argument('--lr', type=float, default=lr)  # The learning rate,
    parser.add_argument('--n_hid', type=int, default=n_hid)  # The number of nodes in the hidden layer
    parser.add_argument('--nheads', type=int, default=nheads)  # The number of attention heads
    parser.add_argument('--nlayers', type=int, default=nlayers)  # The number of graph convolution layers
    parser.add_argument('--cell_size', type=int, default=cell_size)  # The feature dimension or size of a single cell
    parser.add_argument('--neighbor', type=int, default=neighbor)  # The number of neighboring nodes considered in the graph convolution
    parser.add_argument('--egrn', type=bool, default=egrn) 
    parser.add_argument('--output_file', type=str, default=output_file)
    
    # Return all parameters as a tuple
    args = parser.parse_args([])
    return (args.output_file, args.labsm, args.lr, args.wd, 
            args.n_hid, args.nheads, args.nlayers, args.cell_size, 
            args.neighbor, args.egrn)


# Check if the folder exists, and create it if it does not
def ensure_dir(file_path):
    if not os.path.exists(file_path):
        os.makedirs(file_path)


from bioservices import KEGG

def get_cancer_metastasis_genes():
    kegg = KEGG()

    # Define cancer metastasis pathways
    cancer_metastasis_pathways = {
        'VEGF signaling pathway': 'hsa04370',
        'Focal adhesion': 'hsa04510',
        'Pathways in cancer': 'hsa05200',
        'PI3K-Akt signaling pathway': 'hsa04151',
        'MAPK signaling pathway': 'hsa04010',
        "Hippo signaling pathway": "hsa04390",
        "ECM-receptor interaction": "hsa04512"
    }

    pathway_genes = {}

    # Iterate over each pathway
    for pathway_name, pathway_id in cancer_metastasis_pathways.items():
        # Retrieve pathway information
        pathway_info = kegg.get(pathway_id)
        parsed_pathway = kegg.parse(pathway_info)

        # Get and store the gene list
        genes = parsed_pathway['GENE']
        gene_symbols = []
        for gene_id, gene_info in genes.items():
            # Get the gene symbol
            gene_symbol = gene_info.split(' ')[0].split(';')[0]
            gene_symbols.append(gene_symbol)

        pathway_genes[pathway_name] = gene_symbols

    return pathway_genes
    

def initial_clustering(RNA_matrix, custom_n_neighbors=None, n_pcs=40, custom_resolution=None, use_rep=None, random_seed=0):
    print(
        '\tWhen the number of cells is less than or equal to 500, it is recommended to set the resolution value to 0.2.')
    print('\tWhen the number of cells is within the range of 500 to 5000, the resolution value should be set to 0.5.')
    print('\tWhen the number of cells is greater than 5000, the resolution value should be set to 0.8.')

    def segment_function(x):
        if x <= 500:
            return 0.2, 5
        elif x <= 5000:
            return 0.5, 10
        else:
            return 0.8, 15

    adata = ad.AnnData(RNA_matrix.transpose(), dtype='int32')

    # If the user did not provide a custom resolution or n_neighbors value, use the values calculated by segment_function
    if custom_resolution is None or custom_n_neighbors is None:
        resolution, n_neighbors = segment_function(adata.shape[0])
    else:
        resolution = custom_resolution
        n_neighbors = custom_n_neighbors

    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    
    # Use the user-provided embedding if available, otherwise use n_pcs
    if use_rep is not None:
        adata.obsm['use_rep']=use_rep
        sc.pp.neighbors(adata, use_rep='use_rep', n_neighbors=n_neighbors)
    else:
        sc.pp.neighbors(adata, n_neighbors=n_neighbors, n_pcs=n_pcs)

    sc.tl.leiden(adata, resolution, random_state=random_seed)
    return adata.obs['leiden']
    
    
def subgraph_1(graph, seed, n_neighbors, node_sele_prob):
    total_matrix_size = 1 + np.cumprod(n_neighbors).sum()  # Number of nodes in the subgraph
    picked_nodes = {seed}  # One node in the batch
    last_layer_nodes = {seed}

    # Number of nodes selected in each layer. Initially, only the seed node is selected.
    to_pick = 1
    for n_neighbors_current in n_neighbors:  # Current layer neighbors
        to_pick = to_pick * n_neighbors_current
        neighbors = graph[list(last_layer_nodes), :].nonzero()[1]  # Find neighbors of last_layer_nodes

        neighbors_prob = node_sele_prob[list(neighbors)]
        neighbors = list(set(neighbors))  # Make all nodes from the last layer part of the neighbors set
        n_neigbors_real = min(
            to_pick,
            len(neighbors))  # Handle the case where the required number of neighbors is less than the actual number of neighbors
        if len(neighbors_prob) == 0:
            continue
        last_layer_nodes = set(
            np.random.choice(neighbors, n_neigbors_real, replace=False,
                             p=softmax(neighbors_prob)))  # Select non-repeated nodes from neighbors
        picked_nodes |= last_layer_nodes  # Update picked_nodes as last_layer_nodes ∪ picked_nodes
    indices = list(sorted(picked_nodes - {seed}))
    return indices


def softmax(x):
    return (np.exp(x) / np.exp(x).sum())

def subgraph_2(graph, seed, n_neighbors):
    picked_nodes = {seed}
    last_layer_nodes = {seed}

    for n_neighbors_current in n_neighbors:
        neighbors = graph[list(last_layer_nodes), :].nonzero()[1]
        neighbors = list(set(neighbors))
        n_neigbors_real = min(n_neighbors_current, len(neighbors))

        if len(neighbors) == 0:
            continue

        last_layer_nodes = set(np.random.choice(neighbors, n_neigbors_real, replace=False))
        picked_nodes |= last_layer_nodes

    indices = list(sorted(picked_nodes - {seed}))
    return indices

def batch_process_1(args):
    i, node_ids, RNA_matrix, neighbor, cell_size = args
    gene_indices_all = []
    dic = {}

    start_index = i * cell_size
    end_index = min(start_index + cell_size, len(node_ids))

    for node_index in node_ids[start_index:end_index]:
        rna_ = RNA_matrix[:, node_index].todense()
        rna_[rna_ < 5] = 0
        gene_indices = subgraph_1(RNA_matrix.transpose(), node_index, neighbor, np.squeeze(np.array(np.log(rna_ + 1))))
        dic[node_index] = {'g': gene_indices}
        gene_indices_all.extend(gene_indices)

    gene_indices_all = list(set(gene_indices_all))
    node_indices_all = node_ids[start_index:end_index]
    h = {'gene_index': gene_indices_all, 'cell_index': node_indices_all}
    return (h, dic)

def batch_process_2(args):
    i, node_ids, RNA_matrix, neighbor, cell_size = args
    gene_indices_all = []
    dic = {}

    start_index = i * cell_size
    end_index = min(start_index + cell_size, len(node_ids))

    for node_index in node_ids[start_index:end_index]:
        gene_indices = subgraph_2(RNA_matrix.transpose(), node_index, neighbor)
        dic[node_index] = {'g': gene_indices}
        gene_indices_all.extend(gene_indices)

    gene_indices_all = list(set(gene_indices_all))
    node_indices_all = node_ids[start_index:end_index]
    h = {'gene_index': gene_indices_all, 'cell_index': node_indices_all}
    return (h, dic)


def batch_select_whole(RNA_matrix, label, neighbor=[20], cell_size=30):
    print('Partitioning data into batches based on sample type.')
    dic = {}

    # Randomly shuffle cell indices
    shuffled_indices = np.random.choice(len(label), size=len(label), replace=False)

    # Create a dictionary mapping shuffled indices to original indices
    index_mapping = {new_idx: original_idx for original_idx, new_idx in enumerate(shuffled_indices)}

    # Get the corresponding labels based on the shuffled indices
    shuffled_labels = [label[index_mapping[i]] for i in range(len(label))]

    # Process Tumor and Lymph Node samples separately
    tumor_node_ids = [index_mapping[i] for i, l in enumerate(shuffled_labels) if l == 'Tumor']
    lymph_node_ids = [index_mapping[i] for i, l in enumerate(shuffled_labels) if l == 'Lymph Node']

    n_batch_tumor = math.ceil(len(tumor_node_ids) / cell_size)
    n_batch_lymph_node = math.ceil(len(lymph_node_ids) / cell_size)

    with mp.Pool(processes=48) as pool:
        # Process Tumor samples with a progress bar
        tasks_tumor = [(i, tumor_node_ids, RNA_matrix, neighbor, cell_size) for i in range(n_batch_tumor)]
        results_tumor = list(tqdm(pool.imap_unordered(batch_process_2, tasks_tumor), total=n_batch_tumor, desc="Processing Tumor samples"))

        # Process Lymph Node samples with a progress bar
        tasks_lymph_node = [(i, lymph_node_ids, RNA_matrix, neighbor, cell_size) for i in range(n_batch_lymph_node)]
        results_lymph_node = list(tqdm(pool.imap_unordered(batch_process_1, tasks_lymph_node), total=n_batch_lymph_node, desc="Processing Lymph Node samples"))

    # Merge results
    results = results_tumor + results_lymph_node
    indices_ss = [res[0] for res in results]
    for res in results:
        dic.update(res[1])
    
    all_cell_indices = [index for batch in indices_ss for index in batch['cell_index']]
    # The returned node_ids are original indices
    return indices_ss, all_cell_indices, dic


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from scipy.stats import zscore
from collections import defaultdict


# Define the similarity function
def sim(z_i, z_j):
    # Normalize the feature vectors and calculate cosine similarity
    z_i_norm = F.normalize(z_i, p=2, dim=1)
    z_j_norm = F.normalize(z_j, p=2, dim=1)
    return torch.mm(z_i_norm, z_j_norm.t())

# Define the contrastive loss function
def contrastive_loss(z_i, z_j, z_k, temperature):
    # Calculate similarities for the positive and negative pairs
    pos_sim = sim(z_i, z_j) / temperature
    neg_sim = sim(z_i, z_k) / temperature

    # Compute the loss
    exp_pos_sim = torch.exp(pos_sim)
    exp_neg_sim = torch.exp(neg_sim).sum(dim=1, keepdim=True)  # Sum over the negative samples
    loss = -torch.log(exp_pos_sim / (exp_pos_sim + exp_neg_sim))

    return loss.mean()

def compute_losses(z_lymph, high_metastatic_features, low_metastatic_features, z_other_sampled,
                   temperature_d1=0.1, temperature_d2=0.5, temperature_d3=1.0):
    loss_d1 = contrastive_loss(z_lymph, z_lymph, high_metastatic_features, temperature_d1)
    loss_d2 = contrastive_loss(high_metastatic_features, z_lymph, low_metastatic_features, temperature_d2)
    loss_d3 = contrastive_loss(low_metastatic_features, high_metastatic_features, z_other_sampled, temperature_d3)

    return loss_d1, loss_d2, loss_d3

class PathwayUCellLoss(nn.Module):
    def __init__(self, pathway_genes, gene_name_to_index, device):
        super(PathwayUCellLoss, self).__init__()
        self.pathway_genes = pathway_genes
        self.gene_name_to_index = gene_name_to_index
        self.device = device

    def forward(self, RNA_matrix, labels, sample_type):
        pathway_indices = {pathway: [self.gene_name_to_index[gene] for gene in genes 
                                     if gene in self.gene_name_to_index] 
                           for pathway, genes in self.pathway_genes.items()}

        pathway_scores = defaultdict(list)
        for pathway, indices in pathway_indices.items():
            if indices:
                pathway_expression = RNA_matrix[:, indices]
                pathway_expression_tensor = torch.tensor(pathway_expression.todense(), dtype=torch.float32).to(self.device)
                pathway_scores[pathway] = pathway_expression_tensor.mean(dim=1)

        unique_labels, counts = np.unique(labels.cpu(), return_counts=True)
        underrepresented_labels = unique_labels[counts < 0.05 * len(labels)]
        ucell_loss = 0
        high_metastatic_indices = {}
        low_metastatic_indices = {}
        for label in underrepresented_labels:
            label_mask = np.array((labels == label).cpu())
            tumor_mask = (np.array(sample_type) == 'Tumor') & label_mask
            tumor_mask_label = tumor_mask[label_mask]
            total_pathway_scores = torch.sum(torch.stack([pathway_scores[pathway][label_mask] 
                                                          for pathway in pathway_scores]), dim=0)
            z_scores = zscore(total_pathway_scores.cpu().numpy())
            high_metastatic_indices[label] = np.where((z_scores > 1.64) & tumor_mask_label)[0]
            low_metastatic_indices[label] = np.where((z_scores <= 1.64) & tumor_mask_label)[0]
            ucell_loss += torch.mean(total_pathway_scores)
        
        ucell_loss_tensor = -torch.tensor(ucell_loss, dtype=torch.float32) / len(underrepresented_labels) if underrepresented_labels.size > 0 else torch.tensor(0.0, dtype=torch.float32)
        return ucell_loss_tensor, high_metastatic_indices, low_metastatic_indices
        
        
class LabelSmoothing(nn.Module):
    """NLL loss with label smoothing.
    """

    def __init__(self, smoothing=0.0):
        """Constructor for LabelSmoothing module.
        :param smoothing: Label smoothing factor
        """
        super(LabelSmoothing, self).__init__()
        self.confidence = 1.0 - smoothing
        self.smoothing = smoothing

    def forward(self, x, target):
        logprobs = torch.nn.functional.log_softmax(x, dim=-1)
        nll_loss = -logprobs.gather(dim=-1, index=target.unsqueeze(1))
        nll_loss = nll_loss.squeeze(1)
        smooth_loss = -logprobs.mean(dim=-1)
        loss = self.confidence * nll_loss + self.smoothing * smooth_loss
        return loss.mean()


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
    
setup(
    name="EmitGCL",
    version='0.1.1',
    packages=find_packages(),
    install_requires=[
        'anndata==0.8.0',
        'dill==0.3.4',
        'matplotlib==3.5.1',
        'numpy==1.22.3',
        'pandas==1.4.2',
        'scipy==1.9.1',
        'seaborn==0.11.2',
        'scikit-learn==1.1.2',
        'torch-geometric==2.1.0.post1',
        'bioservices==1.11.2',
        'h5py==3.10.0',
        'torchmetrics==0.9.3',
        'tqdm==4.64.0',
        'scanpy==1.9.1',
        'leidenalg==0.8.10',
        'ipywidgets==8.0.6',
        'upsetplot==0.8.0'
    ],
    extras_require={
        # The following packages are not available on PyPI as binary wheels with CPU support.
        # Users must manually install them using the appropriate command:
        # "torch_sparse": ["torch-sparse==0.6.12+cpu"],
        # "torch_scatter": ["torch-scatter==2.0.9+cpu"],
        # "torch_cluster": ["torch-cluster==1.5.9+cpu"],
    },
    python_requires='>=3.8.0, <3.9',
    description="MarsGT: A Python library for rare cell identification (Internal testing only)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: MIT License"
    ]
)


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import GCNConv, GATConv
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.inits import glorot, uniform
from torch_geometric.utils import softmax as Softmax
from torchmetrics.functional import pairwise_cosine_similarity
import math

class GeneralConv(nn.Module):
    def __init__(self, conv_name, in_hid, out_hid, num_types, num_relations, n_heads, dropout, use_norm=True):
        super(GeneralConv, self).__init__()
        self.conv_name = conv_name
        # Store attention matrix results
        self.res_att = None
        # Select different graph convolution layers based on the conv_name parameter
        if self.conv_name == 'hgt':
            self.base_conv = HGTConv(in_hid, out_hid, num_types, num_relations, n_heads, dropout, use_norm)
        elif self.conv_name == 'gcn':
            self.base_conv = GCNConv(in_hid, out_hid)
        elif self.conv_name == 'gat':
            self.base_conv = GATConv(in_hid, out_hid // n_heads, heads=n_heads)

    def forward(self, meta_xs, node_type, edge_index, edge_type):
        if self.conv_name == 'hgt':
            a = self.base_conv(meta_xs, node_type, edge_index, edge_type)
            self.edge_attention_weights = self.base_conv.edge_attention_weights
            return a
        elif self.conv_name == 'gcn':
            return self.base_conv(meta_xs, edge_index)
        elif self.conv_name == 'gat':
            return self.base_conv(meta_xs, edge_index)
        elif self.conv_name == 'dense_hgt':
            return self.base_conv(meta_xs, node_type, edge_index, edge_type)

# HGTConv is the heterogeneous graph transformation convolution layer
class HGTConv(MessagePassing):
    def __init__(self, in_dim, out_dim, num_types, num_relations, n_heads, dropout=0.2, use_norm=True, distribution='uniform', **kwargs):
        super(HGTConv, self).__init__(node_dim=0, aggr='add', **kwargs)

        self.in_dim = in_dim
        self.out_dim = out_dim
        self.num_types = num_types
        self.num_relations = num_relations
        self.total_rel = num_types * num_relations * num_types
        self.n_heads = n_heads
        self.d_k = out_dim // n_heads
        self.sqrt_dk = math.sqrt(self.d_k)
        self.use_norm = use_norm
        self.distribution = distribution
        self.att = None
        self.res_att = None
        self.res = None

        self.k_linears = nn.ModuleList()
        self.q_linears = nn.ModuleList()
        self.v_linears = nn.ModuleList()
        self.a_linears = nn.ModuleList()
        self.norms = nn.ModuleList()

        for t in range(num_types):
            self.k_linears.append(nn.Linear(in_dim, out_dim))
            self.q_linears.append(nn.Linear(in_dim, out_dim))
            self.v_linears.append(nn.Linear(in_dim, out_dim))
            self.a_linears.append(nn.Linear(out_dim, out_dim))
            if use_norm:
                self.norms.append(nn.LayerNorm(out_dim))

        self.relation_pri = nn.Parameter(torch.ones(num_relations, self.n_heads))
        self.relation_att = nn.Parameter(torch.Tensor(num_relations, n_heads, self.d_k, self.d_k))
        self.relation_msg = nn.Parameter(torch.Tensor(num_relations, n_heads, self.d_k, self.d_k))
        self.skip           = nn.Parameter(torch.ones(num_types))
        self.drop           = nn.Dropout(dropout)

        glorot(self.relation_att)
        glorot(self.relation_msg)

    def _initialize_weights(self):
        for m in self.modules():
            print(m)
            if isinstance(m, nn.Linear):
                if self.distribution == 'uniform':
                    torch.nn.init.xavier_uniform_(m.weight, gain=1)
                if self.distribution == 'normal':
                    torch.nn.init.xavier_normal_(m.weight, gain=1)

    def forward(self, node_inp, node_type, edge_index, edge_type):
        return self.propagate(edge_index, node_inp=node_inp, node_type=node_type, edge_type=edge_type)

    def message(self, edge_index_i, node_inp_i, node_inp_j, node_type_i, node_type_j, edge_type, edge_index):
        '''
            j: source, i: target; <j, i>
        '''
        self.edge_attention_weights = {}
        data_size = edge_index_i.size(0)

        self.res_att = torch.zeros(data_size, self.n_heads).to(node_inp_i.device)
        res_msg = torch.zeros(data_size, self.n_heads, self.d_k).to(node_inp_i.device)

        for source_type in range(self.num_types):
            sb = (node_type_j == int(source_type))
            k_linear = self.k_linears[source_type]
            v_linear = self.v_linears[source_type]
            for target_type in range(self.num_types):
                tb = (node_type_i == int(target_type)) & sb
                q_linear = self.q_linears[target_type]
                for relation_type in range(self.num_relations):
                    idx = (edge_type == int(relation_type)) & tb
                    if idx.sum() == 0:
                        continue

                    target_node_vec = node_inp_i[idx]
                    source_node_vec = node_inp_j[idx]

                    q_mat = q_linear(target_node_vec).view(-1, self.n_heads, self.d_k)
                    k_mat = k_linear(source_node_vec).view(-1, self.n_heads, self.d_k)
                    k_mat = torch.bmm(k_mat.transpose(1, 0), self.relation_att[relation_type]).transpose(1, 0)
                    self.res_att[idx] = (q_mat * k_mat).sum(dim=-1) * self.relation_pri[relation_type] / self.sqrt_dk
                    
                    v_mat = v_linear(source_node_vec).view(-1, self.n_heads, self.d_k)
                    res_msg[idx] = torch.bmm(v_mat.transpose(1, 0), self.relation_msg[relation_type]).transpose(1, 0)
                    
                    attention_weights = (q_mat * k_mat).sum(dim=-1) * self.relation_pri[relation_type] / self.sqrt_dk
#                     print(idx.nonzero().squeeze(1).shape)
#                     print(attention_weights.shape)
                    # Save the attention weights and indices for each edge
                    for index, weight in zip(idx.nonzero().squeeze(1), attention_weights):
                        # Average the attention weights for each head
                        source_node_index = edge_index[0][index]  # Source node
                        target_node_index = edge_index[1][index]  # Target node
                        cell_gene_pair = (int(source_node_index), int(target_node_index))
#                         average_weight = weight.mean(dim=-1).item()
#                         self.edge_attention_weights[cell_gene_pair] = average_weight
                        self.edge_attention_weights[cell_gene_pair] = weight.cpu().detach().numpy()
#         print(self.edge_attention_weights[cell_gene_pair].shape,end=": ")
#         print(self.edge_attention_weights[cell_gene_pair])
                        
        res = res_msg * Softmax(self.res_att, edge_index_i).view(-1, self.n_heads, 1)

        return res.view(-1, self.out_dim)

    def update(self, aggr_out, node_inp, node_type):
        aggr_out = F.gelu(aggr_out)
        res = torch.zeros(aggr_out.size(0), self.out_dim).to(node_inp.device)
        for target_type in range(self.num_types):
            idx = (node_type == int(target_type))
            if idx.sum() == 0:
                continue
            trans_out = self.drop(self.a_linears[target_type](aggr_out[idx]))

            alpha = torch.sigmoid(self.skip[target_type])
            if self.use_norm:
                res[idx] = self.norms[target_type](trans_out * alpha + node_inp[idx] * (1 - alpha))
            else:
                res[idx] = trans_out * alpha + node_inp[idx] * (1 - alpha)
        self.res = res
        return res

    def __repr__(self):
        return '{}(in_dim={}, out_dim={}, num_types={}, num_types={})'.format(
            self.__class__.__name__, self.in_dim, self.out_dim,
            self.num_types, self.num_relations)


import os
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import numpy as np
from tqdm import tqdm
from torch.utils.data import SubsetRandomSampler
from loss_function import *
from conv import *
from scipy.sparse import csc_matrix, csr_matrix
from scipy.io import mmwrite,mmread


class GNN_from_raw(nn.Module):
    def __init__(self, in_dim, n_hid, num_types, num_relations, n_heads, n_layers, dropout=0.2, conv_name='hgt',
                 prev_norm=True, last_norm=True):
        super(GNN_from_raw, self).__init__()
        self.gcs = nn.ModuleList()
        self.num_types = num_types
        self.in_dim = in_dim
        self.n_hid = n_hid
        self.adapt_ws = nn.ModuleList()
        self.drop = nn.Dropout(dropout)
        self.embedding1 = nn.ModuleList()

        # Initialize MLP weight matrices
        for ti in range(num_types):
            self.embedding1.append(nn.Linear(in_dim[ti], 256))

        for t in range(num_types):
            self.adapt_ws.append(nn.Linear(256, n_hid))

        # Initialize graph convolution layers
        for l in range(n_layers - 1):
            self.gcs.append(
                GeneralConv(conv_name, n_hid, n_hid, num_types, num_relations, n_heads, dropout, use_norm=prev_norm))
        self.gcs.append(
            GeneralConv(conv_name, n_hid, n_hid, num_types, num_relations, n_heads, dropout, use_norm=last_norm))

    def encode(self, x, t_id):
        h1 = F.relu(self.embedding1[t_id](x))
        return h1

    def forward(self, node_feature, node_type, edge_index, edge_type):
        node_embedding = []
        for t_id in range(self.num_types):
            node_embedding += list(self.encode(node_feature[t_id], t_id))

        node_embedding = torch.stack(node_embedding)
        # Initialize result matrix
        res = torch.zeros(node_embedding.size(0), self.n_hid).to(node_feature[0].device)

        # Process each node type
        for t_id in range(self.num_types):
            idx = (node_type == int(t_id))
            if idx.sum() == 0:
                continue
            # Update result matrix
            res[idx] = torch.tanh(self.adapt_ws[t_id](node_embedding[idx]))

        # Apply dropout to the result matrix
        meta_xs = self.drop(res)
        del res

        # Initialize a container for the final layer's attention weights
        self.final_layer_attention_weights = {}
        # Iterate through graph convolution layers
        for i, gc in enumerate(self.gcs):
            meta_xs = gc(meta_xs, node_type, edge_index, edge_type)
            # Only store the attention weights from the final layer
            if i == len(self.gcs) - 1:
                self.final_layer_attention_weights = gc.edge_attention_weights

        return meta_xs


class Net(nn.Module):
    def __init__(self, dim_in, dim_out):
        super(Net, self).__init__()
        self.fc1 = nn.Linear(dim_in, dim_out)

    def forward(self, x):
        x = F.relu(self.fc1(x))
        return x
        
        
class NodeDimensionReduction(nn.Module):
    def __init__(self, RNA_matrix, indices, ini_p1, n_hid, n_heads, n_layers, labsm, lr, 
                 wd, device, num_types=2, num_relations=1, epochs=1):
        super(NodeDimensionReduction, self).__init__()
        self.RNA_matrix = RNA_matrix
        self.indices = indices
        self.ini_p1 = ini_p1
        self.in_dim = [RNA_matrix.shape[0], RNA_matrix.shape[1]]
        self.n_hid = n_hid
        self.num_types = num_types
        self.num_relations = num_relations
        self.n_heads = n_heads
        self.n_layers = n_layers
        self.labsm = labsm
        self.lr = lr
        self.wd = wd
        self.device = device
        self.epochs = epochs
        self.LabSm = LabelSmoothing(self.labsm)

        self.gnn = GNN_from_raw(in_dim=self.in_dim,
                                n_hid=self.n_hid,
                                num_types=self.num_types,
                                num_relations=self.num_relations,
                                n_heads=self.n_heads,
                                n_layers=self.n_layers,
                                dropout=0.3).to(self.device)

        self.optimizer = torch.optim.AdamW(self.gnn.parameters(), lr=self.lr, weight_decay=self.wd)
        self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(self.optimizer, 'min', factor=0.5, patience=5,
                                                                    verbose=True)

    def train_model(self, n_batch):
        print('The training process for the NodeDimensionReduction model has started. Please wait.')
        self.subgraph_attention_weights = {}

        for epoch in tqdm(range(self.epochs)):
            total_loss_kl1 = 0
            total_loss_cluster = 0
            all_cell_emb = []

            for batch_id in np.arange(n_batch):
                gene_index = self.indices[batch_id]['gene_index']
                cell_index = self.indices[batch_id]['cell_index']
                gene_feature = self.RNA_matrix[list(gene_index),]
                cell_feature = self.RNA_matrix[:, list(cell_index)].T
                gene_feature = torch.tensor(np.array(gene_feature.todense()), dtype=torch.float32).to(self.device)
                cell_feature = torch.tensor(np.array(cell_feature.todense()), dtype=torch.float32).to(self.device)

                node_feature = [cell_feature, gene_feature]
                gene_cell_sub = self.RNA_matrix[list(gene_index),][:, list(cell_index)]
                gene_cell_edge_index1 = list(np.nonzero(gene_cell_sub)[0] + gene_cell_sub.shape[1]) + list(
                    np.nonzero(gene_cell_sub)[1])
                gene_cell_edge_index2 = list(np.nonzero(gene_cell_sub)[1]) + list(
                    np.nonzero(gene_cell_sub)[0] + gene_cell_sub.shape[1])
                gene_cell_edge_index = torch.LongTensor([gene_cell_edge_index1, gene_cell_edge_index2]).to(self.device)

                edge_index = gene_cell_edge_index
                node_type = torch.LongTensor(np.array(
                    list(np.zeros(len(cell_index))) + list(np.ones(len(gene_index))))).to(self.device)
                edge_type = torch.LongTensor(np.array(list(np.zeros(np.nonzero(gene_cell_sub)[0].shape[0])) + list(
                    np.ones(np.nonzero(gene_cell_sub)[1].shape[0])))).to(self.device)
                l = torch.LongTensor(np.array(self.ini_p1)[[cell_index]]).to(self.device)

                node_rep = self.gnn.forward(node_feature, node_type,
                                            edge_index,
                                            edge_type).to(self.device)
                
                self.subgraph_attention_weights[batch_id] = self.gnn.final_layer_attention_weights
                
                cell_emb = node_rep[node_type == 0]
                gene_emb = node_rep[node_type == 1]
                
                all_cell_emb.append(cell_emb)
                
                decoder1 = torch.mm(gene_emb, cell_emb.t())
                gene_cell_sub = torch.tensor(np.array(gene_cell_sub.todense()), dtype=torch.float32).to(self.device)

                logp_x1 = F.log_softmax(decoder1, dim=-1)
                p_y1 = F.softmax(gene_cell_sub, dim=-1)

                loss_kl1 = F.kl_div(logp_x1, p_y1, reduction='mean')
                loss_cluster = self.LabSm(cell_emb, l)

                batch_loss = loss_kl1 + loss_cluster

                self.optimizer.zero_grad()
                batch_loss.backward()
                self.optimizer.step()

                total_loss_kl1 += loss_kl1.item()
                total_loss_cluster += loss_cluster.item()

            total_loss_kl1 /= n_batch
            total_loss_cluster /= n_batch

            all_cell_emb = torch.cat(all_cell_emb, dim=0)
            all_cell_pre = all_cell_emb.argmax(dim=1)
            # Count the occurrences of each unique value in all_cell_pre
            unique_values, counts = torch.unique(all_cell_pre, return_counts=True)
            # Print the unique values and their counts
            # print("Unique values in all_cell_pre and their counts:")
            # for value, count in zip(unique_values, counts):
                # print(f"Value: {value.item()}, Count: {count.item()}")

            print(f'Epoch {epoch + 1}:')
            print(f'  KL Loss: {total_loss_kl1}')
            print(f'  Cluster Loss: {total_loss_cluster}')

        print('The training for the NodeDimensionReduction model has been completed.')
        return self.gnn
        
        
class EmitGCL(nn.Module):
    def __init__(self, gnn, labsm, n_hid, n_batch, device, lr, wd, pathway_genes, gene_names, sample_type, num_epochs=1):
        super(EmitGCL, self).__init__()
        self.lr = lr
        self.wd = wd
        self.gnn = gnn
        self.n_hid = n_hid
        self.n_batch = n_batch
        self.device = device
        self.num_epochs = num_epochs
        self.net = Net(2 * self.n_hid, self.n_hid).to(self.device)
        self.gnn_optimizer = torch.optim.AdamW(self.gnn.parameters(), lr=self.lr, weight_decay=self.wd)
        self.net_optimizer = torch.optim.AdamW(self.net.parameters(), lr=1e-2)
        self.labsm = labsm
        self.LabSm = LabelSmoothing(self.labsm)
        
        # Initialize PathwayUCellLoss
        gene_name_to_index = {gene: i for i, gene in enumerate(gene_names)}
        self.pathway_loss_fn = PathwayUCellLoss(pathway_genes, gene_name_to_index, device)  

    def forward(self, indices, RNA_matrix, ini_p1, sample_type, nodes_id):
        cluster_l = list()
        cluster_kl_l = list()
        sim_l = list()
        nmi_l = list()
        ini_p1 = np.array(ini_p1)
        
        self.subgraph_attention_weights = {}
        
        for epoch in range(self.num_epochs):
            total_loss_kl1 = 0
            total_loss_cluster = 0
            total_ucell_loss = 0
            total_contrastive_loss = torch.tensor(0.0, device=self.device)  # Ensure the initial value is a tensor
            valid_contrastive_loss_count = 0
            all_cell_emb = []
            
            for batch_id in tqdm(np.arange(self.n_batch)):
                gene_index = indices[batch_id]['gene_index']
                cell_index = indices[batch_id]['cell_index']
                gene_feature = RNA_matrix[list(gene_index),]
                cell_feature = RNA_matrix[:, list(cell_index)].T
                gene_feature = torch.tensor(np.array(gene_feature.todense()), dtype=torch.float32).to(self.device)
                cell_feature = torch.tensor(np.array(cell_feature.todense()), dtype=torch.float32).to(self.device)
                node_feature = [cell_feature, gene_feature]
                gene_cell_sub = RNA_matrix[list(gene_index),][:, list(cell_index)]
                gene_cell_edge_index1 = list(np.nonzero(gene_cell_sub)[0] + gene_cell_sub.shape[1]) + list(
                    np.nonzero(gene_cell_sub)[1])
                gene_cell_edge_index2 = list(np.nonzero(gene_cell_sub)[1]) + list(
                    np.nonzero(gene_cell_sub)[0] + gene_cell_sub.shape[1])
                gene_cell_edge_index = torch.LongTensor([gene_cell_edge_index1, gene_cell_edge_index2]).to(self.device)
                node_type = torch.LongTensor(np.array(
                    list(np.zeros(len(cell_index))) + list(np.ones(len(gene_index))))).to(self.device)
                edge_type = torch.LongTensor(np.array(list(np.zeros(np.nonzero(gene_cell_sub)[0].shape[0])) + list(
                    np.ones(np.nonzero(gene_cell_sub)[1].shape[0])))).to(self.device)
                l = torch.LongTensor(np.array(ini_p1)[[cell_index]]).to(self.device)
                node_rep = self.gnn.forward(node_feature, node_type,
                                            gene_cell_edge_index,
                                            edge_type).to(self.device)
                
                self.subgraph_attention_weights[batch_id] = self.gnn.final_layer_attention_weights
                
                cell_emb = node_rep[node_type == 0]
                gene_emb = node_rep[node_type == 1]

                all_cell_emb.append(cell_emb)

                decoder1 = torch.mm(gene_emb, cell_emb.t())
                gene_cell_sub = torch.tensor(np.array(gene_cell_sub.todense()), dtype=torch.float32).to(self.device)
                logp_x1 = F.log_softmax(decoder1, dim=-1)
                p_y1 = F.softmax(gene_cell_sub, dim=-1)

                loss_kl1 = F.kl_div(logp_x1, p_y1, reduction='mean')
                total_loss_kl1 += loss_kl1
                loss_cluster = self.LabSm(cell_emb, l)
                total_loss_cluster += loss_cluster

            all_cell_emb = torch.cat(all_cell_emb, dim=0)
            all_cell_pre = all_cell_emb.argmax(dim=1)
            
            # Count the occurrences of each unique value in all_cell_pre
            unique_values, counts = torch.unique(all_cell_pre, return_counts=True)

            # Print the unique values and their counts
#             print("Unique values in all_cell_pre and their counts:")
#             for value, count in zip(unique_values, counts):
#                 print(f"Value: {value.item()}, Count: {count.item()}")
    
            ucell_loss, high_metastatic_indices, low_metastatic_indices = self.pathway_loss_fn(RNA_matrix.transpose()[nodes_id,:], all_cell_pre, sample_type)
            total_ucell_loss += ucell_loss
            
            for label in high_metastatic_indices:
                label_mask = (all_cell_pre == label)
                lymph_node_mask = (np.array(sample_type) == 'Lymph Node') & label_mask.cpu().numpy()
                z_lymph = all_cell_emb[lymph_node_mask]
                high_metastatic_features = all_cell_emb[label_mask][high_metastatic_indices[label]]
                low_metastatic_features = all_cell_emb[label_mask][low_metastatic_indices[label]]
                other_cells_mask = ~np.isin(np.arange(len(all_cell_pre)), np.concatenate([high_metastatic_indices[label], low_metastatic_indices[label]]))
                z_other = all_cell_emb[other_cells_mask]
                sample_size = int(len(z_other) * 1.0)
                sampler = torch.utils.data.SubsetRandomSampler(torch.randperm(len(z_other))[:sample_size])
                z_other_sampled = z_other[list(sampler)]
                
                loss_d1,loss_d2,loss_d3 = losses = compute_losses(z_lymph, high_metastatic_features, low_metastatic_features, z_other_sampled)

                if not torch.isnan(loss_d1) and not torch.isnan(loss_d2) and not torch.isnan(loss_d3):
                    total_contrastive_loss += loss_d1 + loss_d2 + loss_d3
                    valid_contrastive_loss_count += 1

            if valid_contrastive_loss_count > 0:
                total_contrastive_loss /= valid_contrastive_loss_count

            # For total_contrastive_loss, add a coefficient of 0.3
            weighted_contrastive_loss = total_contrastive_loss * 0.1
            weighted_ucell_loss = total_ucell_loss * 0.3
            weighted_kl_loss = (total_loss_kl1 / self.n_batch) * 10
            weighted_cluster_loss = (total_loss_cluster / self.n_batch) * 1.0
            
            if total_contrastive_loss > 0: 
                total_loss = weighted_kl_loss + weighted_cluster_loss + weighted_ucell_loss + weighted_contrastive_loss
            else:
                total_loss = weighted_kl_loss + weighted_cluster_loss + weighted_ucell_loss

            print(f'Epoch {epoch + 1}:')
            print(f'  KL Loss: {weighted_kl_loss}')
            print(f'  Cluster Loss: {weighted_cluster_loss}')
            print(f'  UCell Loss: {weighted_ucell_loss}')
            print(f'  Total Contrastive Loss: {weighted_contrastive_loss}')
            print(f'  Total Loss: {total_loss}')  # Add printing of the total loss
            
            self.gnn_optimizer.zero_grad()
            self.net_optimizer.zero_grad()
            total_loss.backward()
            self.gnn_optimizer.step()
            self.net_optimizer.step()

        return self.gnn

    def train_model(self, indices, RNA_matrix, ini_p1, sample_type, nodes_id):
        self.train()
        print('The training process for the EmitGCL model has started. Please wait.')
        EmitGCL_gnn = self.forward(indices, RNA_matrix, ini_p1, sample_type, nodes_id)
        print('The training for the EmitGCL model has been completed.')
        return EmitGCL_gnn
        
        
def EmitGCL_pred(RNA_matrix, EmitGCL_gnn, indices, nheads, nodes_id, cell_size, device, gene_names, node_dim_reduction_model):
    n_batch = len(indices)
    embedding = []
    l_pre = []
    EmitGCL_result = {}
        
    # Assume global counts of cells and genes are known
    num_cells = RNA_matrix.shape[1]  # Total global cell count
    num_genes = RNA_matrix.shape[0]  # Total global gene count

    # Initialize global attention matrices, creating one matrix for each head
    global_attention_matrices = {f'head-{i+1}': torch.zeros(num_genes, num_cells, device=device) for i in range(nheads)}

    with torch.no_grad():
        for batch_id in tqdm(range(n_batch)):
            gene_index = indices[batch_id]['gene_index']
            cell_index = indices[batch_id]['cell_index']
            gene_feature = torch.tensor(np.array(RNA_matrix[list(gene_index),].todense()), dtype=torch.float32).to(device)
            cell_feature = torch.tensor(np.array(RNA_matrix[:, list(cell_index)].T.todense()), dtype=torch.float32).to(device)
            node_feature = [cell_feature, gene_feature]
            
            gene_cell_sub = RNA_matrix[list(gene_index),][:, list(cell_index)]
            gene_cell_edge_index = torch.LongTensor([np.nonzero(gene_cell_sub)[0] + gene_cell_sub.shape[1], np.nonzero(gene_cell_sub)[1]]).to(device)
            
            edge_index = gene_cell_edge_index
            node_type = torch.LongTensor(np.array(list(np.zeros(len(cell_index))) + list(np.ones(len(gene_index))))).to(device)
            edge_type = torch.LongTensor(np.array(list(np.zeros(gene_cell_edge_index.shape[1])))).to(device)
            
            node_rep = EmitGCL_gnn.forward(node_feature, node_type, edge_index, edge_type).to(device)
            
            # Retrieve current subgraph's attention weights
            attention_weights = node_dim_reduction_model.subgraph_attention_weights[batch_id]
            
            # Get global cell and gene indices for the current subgraph
            global_gene_indices = indices[batch_id]['gene_index']
            global_cell_indices = indices[batch_id]['cell_index']
            
            # Iterate through each edge in the current subgraph and update global attention matrices
            for (local_gene_idx, local_cell_idx), weights_vector in attention_weights.items():
                global_gene_idx = global_gene_indices[local_gene_idx-len(indices[batch_id]['cell_index'])]
                global_cell_idx = global_cell_indices[local_cell_idx]

                # Update global attention matrices for each head
                for head_index in range(nheads):
                    head_key = f'head-{head_index+1}'
                    weight = weights_vector[head_index]
                    # Correctly convert numpy.float32 to torch.FloatTensor and assign
                    global_attention_matrices[head_key][global_gene_idx, global_cell_idx] = torch.tensor(weight, device=device)


            cell_emb = node_rep[node_type == 0]
            gene_emb = node_rep[node_type == 1]

            if device.type == "cuda":
                cell_emb = cell_emb.cpu()
            embedding.append(cell_emb.detach().numpy())

            cell_pre = list(cell_emb.argmax(dim=1).detach().numpy())
            l_pre.extend(cell_pre)

    cell_embedding = np.vstack(embedding)
    cell_clu = np.array(l_pre)
    
    # Create a mapping from the original indices to the shuffled indices
    index_mapping = {original_index: node_id for original_index, node_id in enumerate(list(nodes_id))}

    # Reorder cell_clu
    sorted_cell_clu = [None] * len(cell_clu)
    for original_index, node_id in index_mapping.items():
        sorted_cell_clu[node_id] = cell_clu[original_index]

    # Reorder cell_embedding
    sorted_cell_embedding = np.zeros_like(cell_embedding)
    for original_index, node_id in index_mapping.items():
        sorted_cell_embedding[node_id, :] = cell_embedding[original_index, :]


    EmitGCL_result = {'pred_label': sorted_cell_clu, 'cell_embedding': sorted_cell_embedding, 'attention_weights': global_attention_matrices}
    return EmitGCL_result
    
    
# Save a single attention matrix in .mtx format
def save_attention_matrix_to_mtx(tensor_matrix, save_path, file_name):
    # Convert PyTorch tensor to numpy array
    if isinstance(tensor_matrix, torch.Tensor):
        tensor_matrix = tensor_matrix.cpu().numpy()
    # Convert numpy array to CSR format sparse matrix
    sparse_matrix = csr_matrix(tensor_matrix)
    # Save as .mtx format using mmwrite
    mmwrite(os.path.join(save_path, f"{file_name}.mtx"), sparse_matrix)
    
    
def save_result(EmitGCL_result, output_file, attention_file, save_attention=True):
    """
    Save the results of EmitGCL including attention matrices and other data.
    
    Parameters:
    - EmitGCL_result: dict containing the results of the EmitGCL model, including 'pred_label', 'cell_embedding', and 'attention_weights'.
    - output_file: str, the file path to save the results.
    - attention_file: str, the file path to save the attention matrices.
    - save_attention: bool, whether to save the attention matrices.
    """
    
    # Save attention matrices if save_attention is True
    if save_attention:
        attention_matrices = EmitGCL_result['attention_weights']
        # Iterate through the dictionary and save each attention matrix
        for head, matrix in attention_matrices.items():
            save_attention_matrix_to_mtx(matrix, attention_file, head)
    
    # Save additional results to .npy files
    np.save(output_file + "pred.npy", EmitGCL_result['pred_label'])
    np.save(output_file + "cell_embedding.npy", EmitGCL_result['cell_embedding'])
